package dev.jonamireh.processor

import com.google.devtools.ksp.KspExperimental
import com.google.devtools.ksp.processing.KSPLogger
import com.google.devtools.ksp.processing.Resolver
import com.google.devtools.ksp.processing.SymbolProcessor
import com.google.devtools.ksp.processing.SymbolProcessorEnvironment
import com.google.devtools.ksp.processing.SymbolProcessorProvider
import com.google.devtools.ksp.symbol.KSAnnotated
import com.google.devtools.ksp.symbol.KSClassDeclaration
import com.google.devtools.ksp.symbol.KSPropertyDeclaration
import com.google.devtools.ksp.symbol.KSVisitorVoid
import com.google.devtools.ksp.validate

class JvmNameProcessorProvider : SymbolProcessorProvider {
    override fun create(environment: SymbolProcessorEnvironment): SymbolProcessor =
        JvmNameProcessor(environment.logger)
}

class JvmNameProcessor(private val logger: KSPLogger) : SymbolProcessor {

    override fun process(resolver: Resolver): List<KSAnnotated> {
        val declarations = resolver.getAllFiles().asSequence().flatMap { it.declarations  }
        val notProcessed = declarations
            .filter { !it.validate() }

        val propertyVisitor = PropertyVisitor(resolver, logger)

        declarations
            .filter { it.validate() }
            .forEach { it.accept(propertyVisitor, Unit) }

        return notProcessed.toList()
    }

    @OptIn(KspExperimental::class)
    class PropertyVisitor(private val resolver: Resolver, private val logger: KSPLogger) : KSVisitorVoid() {
        override fun visitClassDeclaration(classDeclaration: KSClassDeclaration, data: Unit) {
            classDeclaration.getAllProperties().forEach {
                visitPropertyDeclaration(it, Unit)
            }
        }

        override fun visitPropertyDeclaration(property: KSPropertyDeclaration, data: Unit) {
            property.qualifiedName?.let { propertyName ->
                property.getter?.let { getter ->
                    logger.info(
                        "Property getter ${propertyName.asString()} has JVM name of: ${
                            resolver.getJvmName(
                                getter
                            )
                        }", property
                    )
                }
                if(property.isMutable) {
                    property.setter?.let { setter ->
                        logger.info(
                            "Property setter ${propertyName.asString()} has JVM name of: ${
                                resolver.getJvmName(
                                    setter
                                )
                            }", property
                        )
                    }
                }
            }
        }
    }
}